"use strict";
// Copyright IBM Corp. 2018,2019. All Rights Reserved.
// Node module: @loopback/context
// This file is licensed under the MIT License.
// License text available at https://opensource.org/licenses/MIT
Object.defineProperty(exports, "__esModule", { value: true });
const metadata_1 = require("@loopback/metadata");
const binding_1 = require("./binding");
const keys_1 = require("./keys");
/**
 * Metadata key for binding metadata
 */
exports.BINDING_METADATA_KEY = metadata_1.MetadataAccessor.create('binding.metadata');
/**
 * Check if a class implements `Provider` interface
 * @param cls - A class
 */
function isProviderClass(cls) {
    return cls && typeof cls.prototype.value === 'function';
}
exports.isProviderClass = isProviderClass;
/**
 * A factory function to create a template function to bind the target class
 * as a `Provider`.
 * @param target - Target provider class
 */
function asProvider(target) {
    return binding => binding.toProvider(target).tag(keys_1.ContextTags.PROVIDER, {
        [keys_1.ContextTags.TYPE]: keys_1.ContextTags.PROVIDER,
    });
}
exports.asProvider = asProvider;
/**
 * A factory function to create a template function to bind the target class
 * as a class or `Provider`.
 * @param target - Target class, which can be an implementation of `Provider`
 */
function asClassOrProvider(target) {
    // Add a template to bind to a class or provider
    return binding => {
        if (isProviderClass(target)) {
            asProvider(target)(binding);
        }
        else {
            binding.toClass(target);
        }
    };
}
exports.asClassOrProvider = asClassOrProvider;
/**
 * Convert binding scope and tags as a template function
 * @param scopeAndTags - Binding scope and tags
 */
function asBindingTemplate(scopeAndTags) {
    return binding => {
        if (scopeAndTags.scope) {
            binding.inScope(scopeAndTags.scope);
        }
        if (scopeAndTags.tags) {
            if (Array.isArray(scopeAndTags.tags)) {
                binding.tag(...scopeAndTags.tags);
            }
            else {
                binding.tag(scopeAndTags.tags);
            }
        }
    };
}
exports.asBindingTemplate = asBindingTemplate;
/**
 * Get binding metadata for a class
 * @param target - The target class
 */
function getBindingMetadata(target) {
    return metadata_1.MetadataInspector.getClassMetadata(exports.BINDING_METADATA_KEY, target);
}
exports.getBindingMetadata = getBindingMetadata;
/**
 * A binding template function to delete `name` and `key` tags
 */
function removeNameAndKeyTags(binding) {
    if (binding.tagMap) {
        delete binding.tagMap.name;
        delete binding.tagMap.key;
    }
}
exports.removeNameAndKeyTags = removeNameAndKeyTags;
/**
 * Get the binding template for a class with binding metadata
 *
 * @param cls - A class with optional `@bind`
 */
function bindingTemplateFor(cls) {
    const spec = getBindingMetadata(cls);
    const templateFunctions = (spec && spec.templates) || [
        asClassOrProvider(cls),
    ];
    return binding => {
        for (const t of templateFunctions) {
            binding.apply(t);
        }
        if (spec && spec.target !== cls) {
            // Remove name/key tags inherited from base classes
            binding.apply(removeNameAndKeyTags);
        }
    };
}
exports.bindingTemplateFor = bindingTemplateFor;
exports.DEFAULT_TYPE_NAMESPACES = {
    class: 'classes',
    provider: 'providers',
};
/**
 * Create a binding from a class with decorated metadata. The class is attached
 * to the binding as follows:
 * - `binding.toClass(cls)`: if `cls` is a plain class such as `MyController`
 * - `binding.toProvider(cls)`: it `cls` is a value provider class with a
 * prototype method `value()`
 *
 * @param cls - A class. It can be either a plain class or  a value provider class
 * @param options - Options to customize the binding key
 */
function createBindingFromClass(cls, options = {}) {
    const templateFn = bindingTemplateFor(cls);
    let key = options.key;
    if (!key) {
        key = buildBindingKey(cls, options);
    }
    const binding = binding_1.Binding.bind(key).apply(templateFn);
    if (options.name) {
        binding.tag({ name: options.name });
    }
    if (options.type) {
        binding.tag({ type: options.type }, options.type);
    }
    if (options.defaultScope) {
        binding.applyDefaultScope(options.defaultScope);
    }
    return binding;
}
exports.createBindingFromClass = createBindingFromClass;
/**
 * Find/infer binding key namespace for a type
 * @param type - Artifact type, such as `controller`, `datasource`, or `server`
 * @param typeNamespaces - An object mapping type names to namespaces
 */
function getNamespace(type, typeNamespaces = exports.DEFAULT_TYPE_NAMESPACES) {
    if (type in typeNamespaces) {
        return typeNamespaces[type];
    }
    else {
        // Return the plural form
        return `${type}s`;
    }
}
/**
 * Build the binding key for a class with optional binding metadata.
 * The binding key is resolved in the following steps:
 *
 * 1. Check `options.key`, if it exists, return it
 * 2. Check if the binding metadata has `key` tag, if yes, return its tag value
 * 3. Identify `namespace` and `name` to form the binding key as
 * `<namespace>.<name>`.
 *   - namespace
 *     - `options.namespace`
 *     - `namespace` tag value
 *     - Map `options.type` or `type` tag value to a namespace, for example,
 *       'controller` to 'controller'.
 *   - name
 *     - `options.name`
 *     - `name` tag value
 *     - the class name
 *
 * @param cls - A class to be bound
 * @param options - Options to customize how to build the key
 */
function buildBindingKey(cls, options = {}) {
    const templateFn = bindingTemplateFor(cls);
    // Create a temporary binding
    const bindingTemplate = new binding_1.Binding('template').apply(templateFn);
    // Is there a `key` tag?
    let key = options.key || bindingTemplate.tagMap[keys_1.ContextTags.KEY];
    if (key)
        return key;
    let namespace = options.namespace || bindingTemplate.tagMap[keys_1.ContextTags.NAMESPACE];
    if (!namespace) {
        const namespaces = Object.assign({}, exports.DEFAULT_TYPE_NAMESPACES, options.typeNamespaceMapping);
        // Derive the key from type + name
        let type = options.type || bindingTemplate.tagMap[keys_1.ContextTags.TYPE];
        if (!type) {
            type =
                bindingTemplate.tagNames.find(t => namespaces[t] != null) ||
                    keys_1.ContextTags.CLASS;
        }
        namespace = getNamespace(type, namespaces);
    }
    const name = options.name || bindingTemplate.tagMap[keys_1.ContextTags.NAME] || cls.name;
    key = `${namespace}.${name}`;
    return key;
}
//# sourceMappingURL=binding-inspector.js.map